import argparse
import json
import logging
import textwrap

from medallion import (
    __version__, application_instance, register_blueprints, set_config
)

log = logging.getLogger("medallion")


class NewlinesHelpFormatter(argparse.RawDescriptionHelpFormatter):
    """Custom help formatter to insert newlines between argument help texts.
    """
    def _split_lines(self, text, width):
        text = self._whitespace_matcher.sub(" ", text).strip()
        txt = textwrap.wrap(text, width)
        txt[-1] += "\n"
        return txt


def _get_argparser():
    """Create and return an ArgumentParser for this application."""
    desc = "medallion v{0}".format(__version__)
    parser = argparse.ArgumentParser(
        description=desc,
        formatter_class=NewlinesHelpFormatter,
    )

    parser.add_argument(
        "--host",
        default="127.0.0.1",
        type=str,
        help="The host to listen on.",
    )

    parser.add_argument(
        "--port",
        default=5000,
        type=int,
        help="The port of the web server.",
    )

    parser.add_argument(
        "--debug-mode",
        default=False,
        action="store_true",
        help="If set, start application in debug mode.",
    )

    parser.add_argument(
        "--log-level",
        default="WARN",
        type=str,
        help="The logging output level for medallion.",
        choices=["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"],
    )

    parser.add_argument(
        "CONFIG_PATH",
        metavar="CONFIG_PATH",
        type=str,
        help="The location of the JSON configuration file to use.",
    )

    return parser


def main():
    medallion_parser = _get_argparser()
    medallion_args = medallion_parser.parse_args()
    log.setLevel(medallion_args.log_level)

    with open(medallion_args.CONFIG_PATH, "r") as f:
        configuration = json.load(f)

    set_config(application_instance, "users", configuration)
    set_config(application_instance, "taxii", configuration)
    set_config(application_instance, "backend", configuration)
    register_blueprints(application_instance)

    application_instance.run(
        host=medallion_args.host,
        port=medallion_args.port,
        debug=medallion_args.debug_mode,
    )


if __name__ == "__main__":
    main()



import logging
import uuid

from pymongo import MongoClient
from pymongo.errors import ConnectionFailure, ServerSelectionTimeoutError

from ..common import (
    SessionChecker, create_resource, datetime_to_float, datetime_to_string,
    datetime_to_string_stix, determine_spec_version, determine_version,
    find_version_attribute, float_to_datetime, generate_status,
    generate_status_details, get_custom_headers, get_timestamp,
    parse_request_parameters, string_to_datetime
)
from ..exceptions import MongoBackendError, ProcessingError
from ..filters.mongodb_filter import MongoDBFilter
from .base import Backend

# Module-level logger
log = logging.getLogger(__name__)


def catch_mongodb_error(func):
    """Catch mongodb availability error"""

    def api_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (ConnectionFailure, ServerSelectionTimeoutError) as e:
            raise MongoBackendError("Unable to connect to MongoDB", 500, e)

    return api_wrapper


class MongoBackend(Backend):

    # access control is handled at the views level

    def __init__(self, **kwargs):
        try:
            self.client = MongoClient(kwargs.get("uri"))
            self.pages = {}
            self.timeout = kwargs.get("session_timeout", 30)
        except ConnectionFailure:
            log.error("Unable to establish a connection to MongoDB server {}".format(kwargs.get("uri")))

        checker = SessionChecker(kwargs.get("check_interval", 10), self._pop_expired_sessions)
        checker.start()

    def _process_params(self, filter_args, limit):
        next_id = filter_args.get("next")
        if limit and next_id is None:
            client_params = parse_request_parameters(filter_args)
            record = {"skip": 0, "limit": limit, "args": client_params, "request_time": datetime_to_float(get_timestamp())}
            next_id = str(uuid.uuid4())
            self.pages[next_id] = record
        elif limit and next_id:
            if next_id not in self.pages:
                raise ProcessingError("The server did not understand the request or filter parameters: 'next' not valid", 400)
            client_params = parse_request_parameters(filter_args)
            if self.pages[next_id]["args"] != client_params:
                raise ProcessingError("The server did not understand the request or filter parameters: params changed over subsequent transaction", 400)
            self.pages[next_id]["limit"] = limit
            self.pages[next_id]["request_time"] = datetime_to_float(get_timestamp())
            record = self.pages[next_id]
        else:
            record = {}
        return next_id, record

    def _update_record(self, next_id, count, internal=False):
        more = False
        if next_id:
            if internal is False:
                self.pages[next_id]["skip"] += self.pages[next_id]["limit"]
            if self.pages[next_id]["skip"] >= count:
                self.pages.pop(next_id, None)
                next_id = None
            else:
                more = True
        return next_id, more

    def _validate_object_id(self, manifest_info, collection_id, object_id):
        result = list(manifest_info.find({"_collection_id": collection_id, "id": object_id}).limit(1))
        if len(result) == 0:
            raise ProcessingError("Object '{}' not found".format(object_id), 404)

    def _pop_expired_sessions(self):
        expired_ids = []
        boundary = datetime_to_float(get_timestamp())
        for next_id, record in self.pages.items():
            if boundary - record["request_time"] > self.timeout:
                expired_ids.append(next_id)

        for item in expired_ids:
            self.pages.pop(item)

    def _get_object_manifest(self, api_root, collection_id, filter_args, allowed_filters, limit, internal=False):
        api_root_db = self.client[api_root]
        manifest_info = api_root_db["manifests"]
        next_id, record = self._process_params(filter_args, limit)

        full_filter = MongoDBFilter(
            filter_args,
            {"_collection_id": {"$eq": collection_id}},
            allowed_filters,
            record
        )
        count, objects_found = full_filter.process_filter(
            manifest_info,
            allowed_filters,
            None,
        )

        for obj in objects_found:
            obj["date_added"] = datetime_to_string(float_to_datetime(obj["date_added"]))
            obj["version"] = datetime_to_string_stix(float_to_datetime(obj["version"]))

        next_id, more = self._update_record(next_id, count, internal)
        manifest_resource = create_resource("objects", objects_found, more, next_id)
        if internal:
            return manifest_resource
        else:
            headers = get_custom_headers(manifest_resource)
            return manifest_resource, headers

    @catch_mongodb_error
    def _update_manifest(self, new_obj, api_root, collection_id, obj_version, request_time):
        api_root_db = self.client[api_root]
        manifest_info = api_root_db["manifests"]
        collection_info = api_root_db["collections"]
        media_type_fmt = "application/stix+json;version={}"

        media_type = media_type_fmt.format(determine_spec_version(new_obj))

        # version is a single value now, therefore a new manifest is created always
        manifest_info.insert_one(
            {
                "id": new_obj["id"],
                "date_added": datetime_to_float(request_time),
                "version": datetime_to_float(string_to_datetime(obj_version)),
                "media_type": media_type,
                "_collection_id": collection_id,
                "_type": new_obj["type"],
            },
        )

        # update media_types in collection if a new one is present.
        info = collection_info.find_one({"id": collection_id})
        if media_type not in info["media_types"]:
            info["media_types"].append(media_type)
            collection_info.update_one(
                {"id": collection_id},
                {"$set": {"media_types": info["media_types"]}}
            )

    @catch_mongodb_error
    def server_discovery(self):
        discovery_db = self.client["discovery_database"]
        discovery_info = discovery_db["discovery_information"]
        pipeline = [
            {
                "$lookup": {
                    "from": "api_root_info",
                    "localField": "api_roots",
                    "foreignField": "_name",
                    "as": "_roots",
                },
            },
            {
                "$addFields": {
                    "api_roots": "$_roots._url",
                },
            },
            {
                "$project": {
                    "_roots": 0,
                    "_id": 0,
                }
            }
        ]
        info = discovery_info.aggregate(pipeline).next()
        return info

    @catch_mongodb_error
    def get_collections(self, api_root):
        if api_root not in self.client.list_database_names():
            return None  # must return None, so 404 is raised

        api_root_db = self.client[api_root]
        collection_info = api_root_db["collections"]
        collections = list(collection_info.find({}, {"_id": 0}))
        return create_resource("collections", collections)

    @catch_mongodb_error
    def get_collection(self, api_root, collection_id):
        if api_root not in self.client.list_database_names():
            return None  # must return None, so 404 is raised

        api_root_db = self.client[api_root]
        collection_info = api_root_db["collections"]
        info = collection_info.find_one({"id": collection_id}, {"_id": 0})
        return info

    @catch_mongodb_error
    def get_object_manifest(self, api_root, collection_id, filter_args, allowed_filters, limit):
        return self._get_object_manifest(api_root, collection_id, filter_args, allowed_filters, limit, False)

    @catch_mongodb_error
    def get_api_root_information(self, api_root_name):
        db = self.client["discovery_database"]
        api_root_info = db["api_root_info"]
        info = api_root_info.find_one(
            {"_name": api_root_name},
            {"_id": 0, "_url": 0, "_name": 0}
        )
        return info

    @catch_mongodb_error
    def get_status(self, api_root, status_id):
        api_root_db = self.client[api_root]
        status_info = api_root_db["status"]
        result = status_info.find_one(
            {"id": status_id},
            {"_id": 0}
        )
        return result

    @catch_mongodb_error
    def get_objects(self, api_root, collection_id, filter_args, allowed_filters, limit):
        api_root_db = self.client[api_root]
        objects_info = api_root_db["objects"]
        manifest_info = api_root_db["manifests"]
        next_id, record = self._process_params(filter_args, limit)

        full_filter = MongoDBFilter(
            filter_args,
            {"_collection_id": {"$eq": collection_id}},
            allowed_filters,
            record
        )
        # Note: error handling was not added to following call as mongo will
        # handle (user supplied) filters gracefully if they don't exist
        count, objects_found = full_filter.process_filter(
            objects_info,
            allowed_filters,
            manifest_info
        )

        for obj in objects_found:
            if "modified" in obj:
                obj["modified"] = datetime_to_string_stix(float_to_datetime(obj["modified"]))
            if "created" in obj:
                obj["created"] = datetime_to_string_stix(float_to_datetime(obj["created"]))

        manifest_resource = self._get_object_manifest(api_root, collection_id, filter_args, allowed_filters, limit, True)
        headers = get_custom_headers(manifest_resource)

        next_id, more = self._update_record(next_id, count)
        return create_resource("objects", objects_found, more, next_id), headers

    @catch_mongodb_error
    def add_objects(self, api_root, collection_id, objs, request_time):
        api_root_db = self.client[api_root]
        objects_info = api_root_db["objects"]
        manifest_info = api_root_db["manifests"]
        failed = 0
        succeeded = 0
        pending = 0
        successes = []
        failures = []
        media_fmt = "application/stix+json;version={}"

        try:
            for new_obj in objs["objects"]:
                media_type = media_fmt.format(determine_spec_version(new_obj))
                mongo_query = {"_collection_id": collection_id, "id": new_obj["id"], "media_type": media_type}
                if "modified" in new_obj:
                    mongo_query["version"] = datetime_to_float(string_to_datetime(new_obj["modified"]))
                existing_entry = manifest_info.find_one(mongo_query)
                obj_version = determine_version(new_obj, request_time)
                if existing_entry:
                    status_detail = generate_status_details(
                        new_obj["id"], obj_version,
                        message="Unable to process object because an identical entry already exists in collection '{}'.".format(collection_id),
                    )
                    failures.append(status_detail)
                    failed += 1
                else:
                    new_obj.update({"_collection_id": collection_id})
                    if all(prop not in new_obj for prop in ("modified", "created")):
                        new_obj["_date_added"] = datetime_to_float(string_to_datetime(obj_version))  # Special case for un-versioned objects
                    if "modified" in new_obj:
                        new_obj["modified"] = datetime_to_float(string_to_datetime(new_obj["modified"]))
                    if "created" in new_obj:
                        new_obj["created"] = datetime_to_float(string_to_datetime(new_obj["created"]))
                    objects_info.insert_one(new_obj)
                    self._update_manifest(new_obj, api_root, collection_id, obj_version, request_time)
                    status_detail = generate_status_details(
                        new_obj["id"], obj_version,
                        message="Successfully added object to collection '{}'.".format(collection_id)
                    )
                    successes.append(status_detail)
                    succeeded += 1
        except Exception as e:
            # log.exception(e)
            raise ProcessingError("While processing supplied content, an error occurred", 422, e)

        status = generate_status(
            datetime_to_string(request_time), "complete", succeeded, failed,
            pending, successes=successes, failures=failures,
        )
        api_root_db["status"].insert_one(status)
        status.pop("_id", None)
        return status

    @catch_mongodb_error
    def get_object(self, api_root, collection_id, object_id, filter_args, allowed_filters, limit):
        api_root_db = self.client[api_root]
        objects_info = api_root_db["objects"]
        manifest_info = api_root_db["manifests"]
        # set manually to properly retrieve manifests, and early to not break the pagination checks
        filter_args["match[id]"] = object_id
        next_id, record = self._process_params(filter_args, limit)

        self._validate_object_id(manifest_info, collection_id, object_id)

        full_filter = MongoDBFilter(
            filter_args,
            {"_collection_id": {"$eq": collection_id}, "id": {"$eq": object_id}},
            allowed_filters,
            record
        )
        count, objects_found = full_filter.process_filter(
            objects_info,
            allowed_filters,
            manifest_info
        )

        for obj in objects_found:
            if "modified" in obj:
                obj["modified"] = datetime_to_string_stix(float_to_datetime(obj["modified"]))
            if "created" in obj:
                obj["created"] = datetime_to_string_stix(float_to_datetime(obj["created"]))

        manifest_resource = self._get_object_manifest(api_root, collection_id, filter_args, ("id", "type", "version", "spec_version"), limit, True)
        headers = get_custom_headers(manifest_resource)

        next_id, more = self._update_record(next_id, count)
        return create_resource("objects", objects_found, more, next_id), headers

    @catch_mongodb_error
    def delete_object(self, api_root, collection_id, object_id, filter_args, allowed_filters):
        api_root_db = self.client[api_root]
        objects_info = api_root_db["objects"]
        manifest_info = api_root_db["manifests"]

        self._validate_object_id(manifest_info, collection_id, object_id)

        # Currently it will delete the object and the matching manifest from the backend
        full_filter = MongoDBFilter(
            filter_args,
            {"_collection_id": {"$eq": collection_id}, "id": {"$eq": object_id}},
            allowed_filters,
        )
        count, objects_found = full_filter.process_filter(
            objects_info,
            allowed_filters,
            manifest_info
        )
        if objects_found:
            for obj in objects_found:
                attr = find_version_attribute(obj)
                objects_info.delete_one(
                    {"_collection_id": collection_id, "id": object_id, attr: obj[attr]}
                )
                obj_version = obj.get("modified", obj.get("created", obj.get("_date_added")))
                manifest_info.delete_one(
                    {"_collection_id": collection_id, "id": object_id, "version": obj_version}
                )
        else:
            raise ProcessingError("Object '{}' not found".format(object_id), 404)

    @catch_mongodb_error
    def get_object_versions(self, api_root, collection_id, object_id, filter_args, allowed_filters, limit):
        api_root_db = self.client[api_root]
        manifest_info = api_root_db["manifests"]
        # set manually to properly retrieve manifests, and early to not break the pagination checks
        filter_args["match[id]"] = object_id
        filter_args["match[version]"] = "all"
        next_id, record = self._process_params(filter_args, limit)

        self._validate_object_id(manifest_info, collection_id, object_id)

        full_filter = MongoDBFilter(
            filter_args,
            {"_collection_id": {"$eq": collection_id}, "id": {"$eq": object_id}},
            allowed_filters,
            record
        )
        count, manifests_found = full_filter.process_filter(
            manifest_info,
            allowed_filters,
            None,
        )

        manifest_resource = self._get_object_manifest(api_root, collection_id, filter_args, ("id", "type", "version", "spec_version"), limit, True)
        headers = get_custom_headers(manifest_resource)

        manifests_found = list(map(lambda x: datetime_to_string_stix(float_to_datetime(x["version"])), manifests_found))
        next_id, more = self._update_record(next_id, count)
        return create_resource("versions", manifests_found, more, next_id), headers



import pymongo

from ..common import datetime_to_float, string_to_datetime
from .basic_filter import BasicFilter


class MongoDBFilter(BasicFilter):

    def __init__(self, filter_args, basic_filter, allowed, record=None):
        super(MongoDBFilter, self).__init__(filter_args)
        self.basic_filter = basic_filter
        self.full_query = self._query_parameters(allowed)
        self.record = record

    def _query_parameters(self, allowed):
        parameters = self.basic_filter
        if self.filter_args:
            match_type = self.filter_args.get("match[type]")
            if match_type and "type" in allowed:
                types_ = match_type.split(",")
                if len(types_) == 1:
                    parameters["_type"] = {"$eq": types_[0]}
                else:
                    parameters["_type"] = {"$in": types_}
            match_id = self.filter_args.get("match[id]")
            if match_id and "id" in allowed:
                ids_ = match_id.split(",")
                if len(ids_) == 1:
                    parameters["id"] = {"$eq": ids_[0]}
                else:
                    parameters["id"] = {"$in": ids_}
        return parameters

    def process_filter(self, data, allowed, manifest_info):
        match_filter = {
            "$match": {
                "$and": [self.full_query],
                "$comment": "Step #1: Match against an object/manifest id or type present in a collection.",
            },
        }
        pipeline = [match_filter]

        # create added_after filter
        added_after_date = self.filter_args.get("added_after")
        if added_after_date:
            added_after_timestamp = datetime_to_float(string_to_datetime(added_after_date))
            date_filter = {
                "$match": {
                    "date_added": {"$gt": added_after_timestamp},
                    "$comment": "Step #2: If added_after is provided, remove all objects/manifests older than the provided time",
                }
            }
            pipeline.append(date_filter)

        # create spec_version filter. when no filter is provided only latest is considered.
        if "spec_version" in allowed:
            match_spec_version = self.filter_args.get("match[spec_version]")
            if match_spec_version:
                spec_versions = match_spec_version.split(",")
                media_fmt = "application/stix+json;version={}"
                if len(spec_versions) == 1:
                    pipeline[0]["$match"]["$and"].append({"media_type": {"$eq": media_fmt.format(spec_versions[0])}})
                else:
                    pipeline[0]["$match"]["$and"].append({"media_type": {"$in": [media_fmt.format(x) for x in spec_versions]}})
            else:
                pipeline.append({"$group": {"_id": "$id", "objs": {"$push": "$$ROOT"}, "media_types": {"$push": "$media_type"}}})
                pipeline.append({"$sort": {"media_types": pymongo.ASCENDING}})
                pipeline.append({"$addFields": {"media_types": {"$arrayElemAt": ["$media_types", -1]}}})
                redact_objects = {
                    "$redact": {
                        "$cond": {
                            "if": {"$ne": ["$objs.media_type", "$media_types"]},
                            "then": "$$KEEP",
                            "else": "$$PRUNE",
                        }
                    }
                }
                pipeline.append(redact_objects)
                pipeline.append({"$unwind": "$objs"})
                pipeline.append({"$replaceRoot": {"newRoot": "$objs"}})

        # create version filter
        if "version" in allowed:
            match_version = self.filter_args.get("match[version]")
            if not match_version:
                match_version = "last"
            if "all" not in match_version:
                actual_dates = [datetime_to_float(string_to_datetime(x)) for x in match_version.split(",") if (x != "first" and x != "last")]
                # If specific dates have been selected, then we add these to the $match criteria
                # created from the self.full_query at the beginning of this method. This will make
                # sure we can pick the correct manifests even if `added_after` later modifies this results.
                if len(actual_dates) > 0:
                    pipeline[0]["$match"]["$and"].append({"version": {"$in": actual_dates}})

                pipeline.append({"$sort": {"version": pymongo.ASCENDING}})
                pipeline.append({"$group": {"_id": "$id", "versions": {"$push": "$$ROOT"}}})

                # The versions array in the mongodb document is ordered oldest to newest, so the 'last'
                # (most recent date) is in last position in the list and the oldest 'first' is in
                # the first position.
                version_selector = []
                if "last" in match_version:
                    version_selector.append({"$arrayElemAt": ["$versions", -1]})
                if "first" in match_version:
                    version_selector.append({"$arrayElemAt": ["$versions", 0]})
                for d in actual_dates:
                    version_selector.append({"$arrayElemAt": ["$versions", {"$indexOfArray": ["$versions", d]}]})
                pipeline.append({"$addFields": {"versions": version_selector}})

                # denormalize the embedded objects, remove duplicates and replace the document root. Could be improved
                pipeline.append({"$unwind": "$versions"})
                pipeline.append({"$replaceRoot": {"newRoot": "$versions"}})
                pipeline.append({"$group": {"_id": "$$ROOT"}})
                pipeline.append({"$replaceRoot": {"newRoot": "$_id"}})
        pipeline.append({"$sort": {"date_added": pymongo.ASCENDING}})

        if data.name == "manifests":
            # Project the final results
            project_results = {"$project": {"_id": 0, "_collection_id": 0, "_type": 0}}
            pipeline.append(project_results)

            count = self.get_result_count(pipeline, data)
            self.add_pagination_operations(pipeline)
            cursor = data.aggregate(pipeline)
            results = list(cursor)
        else:
            # Join the filtered manifest(s) to the objects collection
            join_objects = {
                "$lookup": {
                    "from": "objects",
                    "localField": "id",
                    "foreignField": "id",
                    "as": "obj",
                },
            }
            pipeline.append(join_objects)

            # Copy the filtered version list to the embedded object document
            add_versions = {
                "$addFields": {"obj.version": "$version"},
            }
            pipeline.append(add_versions)

            # Copy the media_type we a looking for into the embedded object document
            add_media_type = {
                "$addFields": {"obj.media_type": "$media_type"},
            }
            pipeline.append(add_media_type)

            # denormalize the embedded objects and replace the document root
            pipeline.append({"$unwind": "$obj"})
            pipeline.append({"$replaceRoot": {"newRoot": "$obj"}})

            # Redact the result set removing objects where the modified date is not in
            # the version field and the object isn't in the correct collection.
            # The collection filter is required because the join between manifests and objects
            # does not include collection_id.
            #
            col_id = self.full_query["_collection_id"]["$eq"]
            redact_objects = {
                "$redact": {
                    "$cond": {
                        "if": {
                            "$and": [
                                {"$eq": ["$_collection_id", col_id]},
                                {
                                    "$switch": {
                                        "branches": [
                                            {"case": {"$eq": ["$modified", "$version"]}, "then": True},
                                            {"case": {"$and": [
                                                {"$eq": ["$created", "$version"]}, {"$not": ["$modified"]}
                                            ]}, "then": True},
                                            {"case": {"$eq": ["$_date_added", "$version"]}, "then": True},
                                        ],
                                        "default": False,
                                    },
                                },
                                {
                                    "$switch": {
                                        "branches": [
                                            {"case": {"$eq": ["$spec_version", {"$substrBytes": ["$media_type", 30, 4]}]}, "then": True},
                                            {"case": {"$and": [
                                                {"$eq": ["2.0", {"$substrBytes": ["$media_type", 30, 4]}]}, {"$not": ["$spec_version"]}
                                            ]}, "then": True},
                                        ],
                                        "default": False,
                                    },
                                }
                            ],
                        },
                        "then": "$$KEEP",
                        "else": "$$PRUNE",
                    },
                },
            }
            pipeline.append(redact_objects)

            # denormalize the embedded objects, remove duplicates and replace the document root. Could be improved
            pipeline.append({"$group": {"_id": "$$ROOT"}})
            pipeline.append({"$replaceRoot": {"newRoot": "$_id"}})
            pipeline.append({"$sort": {"_date_added": pymongo.ASCENDING, "created": pymongo.ASCENDING, "modified": pymongo.ASCENDING}})

            # Project the final results
            project_results = {"$project": {"version": 0, "media_type": 0, "_id": 0, "_collection_id": 0, "_date_added": 0}}
            pipeline.append(project_results)

            count = self.get_result_count(pipeline, manifest_info)
            self.add_pagination_operations(pipeline)

            cursor = manifest_info.aggregate(pipeline)
            results = list(cursor)

        return count, results

    def add_pagination_operations(self, pipeline):
        if self.record:
            pipeline.append({"$skip": self.record["skip"]})
            pipeline.append({"$limit": self.record["limit"]})

    def get_result_count(self, pipeline, data):
        count_pipeline = list(pipeline)
        count_pipeline.append({"$count": "total"})
        count_result = list(data.aggregate(count_pipeline))

        if len(count_result) == 0:
            # No results
            return 0

        count = count_result[0]["total"]
        return count



import copy
import io
import json
import uuid

from six import string_types

from ..common import (
    SessionChecker, create_resource, datetime_to_float, datetime_to_string,
    determine_spec_version, determine_version, find_att, generate_status,
    generate_status_details, get_timestamp, iterpath
)
from ..exceptions import ProcessingError
from ..filters.basic_filter import BasicFilter
from .base import Backend


def remove_hidden_field(objs):
    for obj in objs:
        if "_date_added" in obj:
            del obj["_date_added"]


def find_headers(headers, manifest, obj):
    obj_time = find_att(obj)
    for man in manifest:
        if man["id"] == obj["id"] and obj_time == find_att(man):
            if len(headers) == 0:
                headers["X-TAXII-Date-Added-First"] = man["date_added"]
            else:
                headers["X-TAXII-Date-Added-Last"] = man["date_added"]


class MemoryBackend(Backend):

    # access control is handled at the views level

    def __init__(self, **kwargs):
        if kwargs.get("filename"):
            self.load_data_from_file(kwargs.get("filename"))
        else:
            self.data = {}
        self.next = {}
        self.timeout = kwargs.get("session_timeout", 30)

        checker = SessionChecker(kwargs.get("check_interval", 10), self._pop_expired_sessions)
        checker.start()

    def set_next(self, objects, args):
        u = str(uuid.uuid4())
        if "limit" in args:
            del args["limit"]
        for arg in args:
            new_list = args[arg].split(',')
            new_list.sort()
            args[arg] = new_list
        d = {"objects": objects, "args": args, "request_time": datetime_to_float(get_timestamp())}
        self.next[u] = d
        return u

    def get_next(self, filter_args, allowed, manifest, lim):
        n = filter_args["next"]
        if n in self.next:
            for arg in filter_args:
                new_list = filter_args[arg].split(',')
                new_list.sort()
                filter_args[arg] = new_list
            del filter_args["next"]
            del filter_args["limit"]
            if filter_args != self.next[n]["args"]:
                raise ProcessingError("The server did not understand the request or filter parameters: params changed over subsequent transaction", 400)
            t = self.next[n]["objects"]
            length = len(self.next[n]["objects"])
            headers = {}
            ret = []
            if length <= lim:
                limit = length
                more = False
                nex = None
            else:
                limit = lim
                more = True

            for i in range(0, limit):
                x = t.pop(0)
                ret.append(x)
                if len(headers) == 0:
                    find_headers(headers, manifest, x)
                if i == limit - 1:
                    find_headers(headers, manifest, x)
            if not more:
                self.next.pop(n)
            else:
                nex = n

            return ret, more, headers, nex
        else:
            raise ProcessingError("The server did not understand the request or filter parameters: 'next' not valid", 400)

    def _pop_expired_sessions(self):
        expired_ids = []
        boundary = datetime_to_float(get_timestamp())
        for next_id, record in self.next.items():
            if boundary - record["request_time"] > self.timeout:
                expired_ids.append(next_id)

        for item in expired_ids:
            self.next.pop(item)

    def load_data_from_file(self, filename):
        if isinstance(filename, string_types):
            with io.open(filename, "r", encoding="utf-8") as infile:
                self.data = json.load(infile)
        else:
            self.data = json.load(filename)

    def save_data_to_file(self, filename, **kwargs):
        """The kwargs are passed to ``json.dump()`` if provided."""
        if isinstance(filename, string_types):
            with io.open(filename, "w", encoding="utf-8") as outfile:
                json.dump(self.data, outfile, **kwargs)
        else:
            json.dump(self.data, filename, **kwargs)

    def _get(self, key):
        for ancestors, item in iterpath(self.data):
            if key in ancestors:
                return item

    def server_discovery(self):
        return self._get("/discovery")

    def _update_manifest(self, new_obj, api_root, collection_id, request_time):
        api_info = self._get(api_root)
        collections = api_info.get("collections", [])
        media_type_fmt = "application/stix+json;version={}"

        for collection in collections:
            if collection_id == collection["id"]:
                version = determine_version(new_obj, request_time)
                request_time = datetime_to_string(request_time)
                media_type = media_type_fmt.format(determine_spec_version(new_obj))

                # version is a single value now, therefore a new manifest is always created
                collection["manifest"].append(
                    {
                        "id": new_obj["id"],
                        "date_added": request_time,
                        "version": version,
                        "media_type": media_type,
                    },
                )

                # if the media type is new, attach it to the collection
                if media_type not in collection["media_types"]:
                    collection["media_types"].append(media_type)

                # quit once you have found the collection that needed updating
                break

    def get_collections(self, api_root):
        if api_root not in self.data:
            return None  # must return None so 404 is raised

        api_info = self._get(api_root)
        collections = copy.deepcopy(api_info.get("collections", []))

        # Remove data that is not part of the response.
        for collection in collections:
            collection.pop("manifest", None)
            collection.pop("responses", None)
            collection.pop("objects", None)
        return create_resource("collections", collections)

    def get_collection(self, api_root, collection_id):
        if api_root not in self.data:
            return None  # must return None so 404 is raised

        api_info = self._get(api_root)
        collections = copy.deepcopy(api_info.get("collections", []))

        for collection in collections:
            if collection_id == collection["id"]:
                collection.pop("manifest", None)
                collection.pop("responses", None)
                collection.pop("objects", None)
                return collection

    def get_object_manifest(self, api_root, collection_id, filter_args, allowed_filters, limit):
        more = False
        n = None
        if api_root in self.data:
            api_info = self._get(api_root)
            collections = api_info.get("collections", [])

            for collection in collections:
                if collection_id == collection["id"]:
                    if "next" in filter_args:
                        manifest = collection.get("manifest", [])
                        manifest, more, headers, n = self.get_next(filter_args, allowed_filters, manifest, limit)
                    else:
                        manifest = collection.get("manifest", [])
                        full_filter = BasicFilter(filter_args)
                        manifest, next_save, headers = full_filter.process_filter(
                            manifest,
                            allowed_filters,
                            None,
                            limit
                        )
                        if len(next_save) != 0:
                            more = True
                            n = self.set_next(next_save, filter_args)
                        break
            return create_resource("objects", manifest, more, n), headers

    def get_api_root_information(self, api_root):
        if api_root in self.data:
            api_info = self._get(api_root)

            if "information" in api_info:
                return api_info["information"]

    def get_status(self, api_root, status_id):
        if api_root in self.data:
            api_info = self._get(api_root)

            for status in api_info.get("status", []):
                if status_id == status["id"]:
                    return status

    def get_objects(self, api_root, collection_id, filter_args, allowed_filters, limit):
        more = False
        n = None
        if api_root in self.data:
            api_info = self._get(api_root)
            collections = api_info.get("collections", [])
            objs = []
            for collection in collections:
                if collection_id == collection["id"]:
                    manifest = collection.get("manifest", [])
                    if "next" in filter_args:
                        objs, more, headers, n = self.get_next(filter_args, allowed_filters, manifest, limit)
                    else:
                        objs = copy.deepcopy(collection.get("objects", []))
                        full_filter = BasicFilter(filter_args)
                        objs, next_save, headers = full_filter.process_filter(
                            objs,
                            allowed_filters,
                            manifest,
                            limit
                        )

                        if len(next_save) != 0:
                            more = True
                            n = self.set_next(next_save, filter_args)
                        break
            remove_hidden_field(objs)
            return create_resource("objects", objs, more, n), headers

    def add_objects(self, api_root, collection_id, objs, request_time):
        if api_root in self.data:
            api_info = self._get(api_root)
            collections = api_info.get("collections", [])
            failed = 0
            succeeded = 0
            pending = 0
            successes = []
            failures = []

            for collection in collections:
                if collection_id == collection["id"]:
                    if "objects" not in collection:
                        collection["objects"] = []
                    try:
                        for new_obj in objs["objects"]:
                            id_and_version_already_present = False
                            for obj in collection["objects"]:
                                if new_obj["id"] == obj["id"]:
                                    if "modified" in new_obj:
                                        if new_obj["modified"] == obj["modified"]:
                                            id_and_version_already_present = True
                                            break
                                    else:
                                        # There is no modified field, so this object is immutable
                                        id_and_version_already_present = True
                                        break
                            if id_and_version_already_present is False:
                                version = determine_version(new_obj, request_time)
                                if "modified" not in new_obj and "created" not in new_obj:
                                    new_obj["_date_added"] = version
                                collection["objects"].append(new_obj)
                                self._update_manifest(new_obj, api_root, collection["id"], request_time)
                                status_details = generate_status_details(
                                    new_obj["id"], version
                                )
                                successes.append(status_details)
                                succeeded += 1
                            else:
                                status_details = generate_status_details(
                                    new_obj["id"], determine_version(new_obj, request_time),
                                    message="Unable to process object",
                                )
                                failures.append(status_details)
                                failed += 1
                    except Exception as e:
                        raise ProcessingError("While processing supplied content, an error occurred", 422, e)

            status = generate_status(
                datetime_to_string(request_time), "complete", succeeded,
                failed, pending, successes=successes,
                failures=failures,
            )
            api_info["status"].append(status)
            return status

    def get_object(self, api_root, collection_id, object_id, filter_args, allowed_filters, limit):
        more = False
        n = None
        if api_root in self.data:
            api_info = self._get(api_root)
            collections = api_info.get("collections", [])
            objs = []
            manifests = []
            for collection in collections:
                if collection_id == collection["id"]:
                    manifests = collection.get("manifest", [])
                    if "next" in filter_args:
                        objs, more, headers, n = self.get_next(filter_args, allowed_filters, manifests, limit)
                    else:
                        for obj in collection.get("objects", []):
                            if object_id == obj["id"]:
                                objs.append(obj)
                        if len(objs) == 0:
                            raise ProcessingError("Object '{}' not found".format(object_id), 404)
                        full_filter = BasicFilter(filter_args)
                        objs, next_save, headers = full_filter.process_filter(
                            objs,
                            allowed_filters,
                            manifests,
                            limit
                        )
                        if len(next_save) != 0:
                            more = True
                            n = self.set_next(next_save, filter_args)
                        break
            remove_hidden_field(objs)
            return create_resource("objects", objs, more, n), headers

    def delete_object(self, api_root, collection_id, obj_id, filter_args, allowed_filters):
        if api_root in self.data:
            api_info = self._get(api_root)
            collections = api_info.get("collections", [])
            objs = []
            manifests = []
            for collection in collections:
                if "id" in collection and collection_id == collection["id"]:
                    coll = collection.get("objects", [])
                    for obj in coll:
                        if obj_id == obj["id"]:
                            objs.append(obj)
                    manifests = collection.get("manifest", [])
                    break

            full_filter = BasicFilter(filter_args)
            objs, nex, headers = full_filter.process_filter(
                objs,
                allowed_filters,
                manifests,
                None
            )

            if len(objs) == 0:
                raise ProcessingError("Object '{}' not found".format(obj_id), 404)

            for obj in objs:
                if obj in coll:
                    coll.remove(obj)
                    obj_time = find_att(obj)
                    for man in manifests:
                        if obj["id"] == man["id"] and obj_time == find_att(man):
                            manifests.remove(man)
                            break

    def get_object_versions(self, api_root, collection_id, object_id, filter_args, allowed_filters, limit):
        more = False
        n = None
        if api_root in self.data:
            api_info = self._get(api_root)
            collections = api_info.get("collections", [])

            objs = []
            for collection in collections:
                if collection_id == collection["id"]:
                    all_manifests = collection.get("manifest", [])
                    if "next" in filter_args:
                        objs, more, headers, n = self.get_next(filter_args, allowed_filters, all_manifests, limit)
                        objs = sorted(map(lambda x: x["version"], objs), reverse=True)
                    else:

                        all_manifests = collection.get("manifest", [])
                        for manifest in all_manifests:
                            if object_id == manifest["id"]:
                                objs.append(manifest)
                        if len(objs) == 0:
                            raise ProcessingError("Object '{}' not found".format(object_id), 404)
                        full_filter = BasicFilter(filter_args)
                        objs, next_save, headers = full_filter.process_filter(
                            objs,
                            allowed_filters,
                            None,
                            limit
                        )
                        if len(next_save) != 0:
                            more = True
                            n = self.set_next(next_save, filter_args)
                        objs = sorted(map(lambda x: x["version"], objs), reverse=True)
                        break
            return create_resource("versions", objs, more, n), headers



# Medallion Custom Exceptions


class MedallionError(Exception):
    """Base error class for Medallion

    Args:
        message (str): specific error message
        root_exception (Exception): Exception instance of root exception
    """
    def __init__(self, message, status, root_exception=None):
        self.message = message
        self.status = status
        self.root_exception = root_exception

    def __str__(self):
        if self.root_exception is not None:
            return "{0.message}. Root exception: {0.root_exception}".format(self)
        else:
            return "{0.message}.".format(self)


class ProcessingError(MedallionError):
    """Internal processing error when processing user supplied data"""
    pass


class BackendError(MedallionError):
    """Medallion data backend error"""
    pass


class MongoBackendError(BackendError):
    """Cannot connect or obtain access to MongoDB backend"""
    pass



import calendar
import datetime as dt
import threading
import uuid

import pytz
from six import iteritems


def create_resource(resource_name, items, more=False, next_id=None):
    """Generates a Resource Object given a resource name."""
    resource = {}
    if items:
        resource[resource_name] = items
    if resource_name == "objects" or resource_name == "versions":
        if next_id and resource:
            resource["next"] = next_id
        if resource:
            resource["more"] = more
    return resource


def determine_version(new_obj, request_time):
    """Grab the modified time if present, if not grab created time,
    if not grab request time provided by server."""
    return new_obj.get("modified", new_obj.get("created", datetime_to_string(request_time)))


def determine_spec_version(obj):
    """Given a STIX 2.x object, determine its spec version."""
    missing = ("created", "modified")
    if all(x not in obj for x in missing):
        # Special case: only SCOs are 2.1 objects and they don't have a spec_version
        # For now the only way to identify one is checking the created and modified
        # are both missing.
        return "2.1"
    return obj.get("spec_version", "2.0")


def get(data, key):
    """Given a dict, loop recursively over the object. Returns the value based on the key match"""
    for ancestors, item in iterpath(data):
        if key in ancestors:
            return item


def iterpath(obj, path=None):
    """
    Generator which walks the input ``obj`` model. Each iteration yields a
    tuple containing a list of ancestors and the property value.

    Args:
        obj: A SDO or SRO object.
        path: None, used recursively to store ancestors.

    Example:
        >>> for item in iterpath(obj):
        >>>     print(item)
        (['type'], 'campaign')
        ...
        (['cybox', 'objects', '[0]', 'hashes', 'sha1'], 'cac35ec206d868b7d7cb0b55f31d9425b075082b')

    Returns:
        tuple: Containing two items: a list of ancestors and the property value.

    """
    if path is None:
        path = []

    for varname, varobj in iter(sorted(iteritems(obj))):
        path.append(varname)
        yield (path, varobj)

        if isinstance(varobj, dict):

            for item in iterpath(varobj, path):
                yield item

        elif isinstance(varobj, list):

            for item in varobj:
                index = "[{0}]".format(varobj.index(item))
                path.append(index)

                yield (path, item)

                if isinstance(item, dict):
                    for descendant in iterpath(item, path):
                        yield descendant

                path.pop()

        path.pop()


def get_timestamp():
    """Get current time with UTC offset"""
    return dt.datetime.now(tz=pytz.UTC)


def datetime_to_string(dttm):
    """Given a datetime instance, produce the string representation
    with microsecond precision"""
    # 1. Convert to timezone-aware
    # 2. Convert to UTC
    # 3. Format in ISO format with microsecond precision

    if dttm.tzinfo is None or dttm.tzinfo.utcoffset(dttm) is None:
        # dttm is timezone-naive; assume UTC
        zoned = pytz.UTC.localize(dttm)
    else:
        zoned = dttm.astimezone(pytz.UTC)
    return zoned.strftime("%Y-%m-%dT%H:%M:%S.%fZ")


def datetime_to_string_stix(dttm):
    """Given a datetime instance, produce the string representation
    with millisecond precision"""
    # 1. Convert to timezone-aware
    # 2. Convert to UTC
    # 3. Format in ISO format with millisecond precision,
    #       except for objects defined with higher precision
    # 4. Add "Z"

    if dttm.tzinfo is None or dttm.tzinfo.utcoffset(dttm) is None:
        # dttm is timezone-naive; assume UTC
        zoned = pytz.UTC.localize(dttm)
    else:
        zoned = dttm.astimezone(pytz.UTC)
    ts = zoned.strftime("%Y-%m-%dT%H:%M:%S")
    ms = zoned.strftime("%f")
    if len(ms[3:].rstrip("0")) >= 1:
        ts = ts + "." + ms + "Z"
    else:
        ts = ts + "." + ms[:3] + "Z"
    return ts


def datetime_to_float(dttm):
    """Given a datetime instance, return its representation as a float"""
    # Based on this solution: https://stackoverflow.com/questions/30020988/python3-datetime-timestamp-in-python2
    if dttm.tzinfo is None:
        return calendar.timegm(dttm.utctimetuple()) + dttm.microsecond / 1e6
    else:
        return (dttm - dt.datetime(1970, 1, 1, tzinfo=pytz.UTC)).total_seconds()


def float_to_datetime(timestamp_float):
    """Given a floating-point number, produce a datetime instance"""
    return dt.datetime.utcfromtimestamp(timestamp_float)


def string_to_datetime(timestamp_string):
    """Convert string timestamp to datetime instance."""
    try:
        return dt.datetime.strptime(timestamp_string, "%Y-%m-%dT%H:%M:%S.%fZ")
    except ValueError:
        return dt.datetime.strptime(timestamp_string, "%Y-%m-%dT%H:%M:%SZ")


def generate_status(
    request_time, status, succeeded, failed, pending,
    successes=None, failures=None, pendings=None,
):
    """Generate Status Resource as defined in TAXII 2.1 section (4.3.1) <link here>`__."""
    status = {
        "id": str(uuid.uuid4()),
        "status": status,
        "request_timestamp": request_time,
        "total_count": succeeded + failed + pending,
        "success_count": succeeded,
        "failure_count": failed,
        "pending_count": pending,
    }

    if successes:
        status["successes"] = successes
    if failures:
        status["failures"] = failures
    if pendings:
        status["pendings"] = pendings

    return status


def generate_status_details(id, version, message=None):
    """Generate Status Details as defined in TAXII 2.1 section (4.3.1) <link here>`__."""
    status_details = {
        "id": id,
        "version": version
    }

    if message:
        status_details["message"] = message

    return status_details


def get_custom_headers(manifest_resource):
    """Generates the X-TAXII-Date-Added headers based on a manifest resource"""
    headers = {}

    times = sorted(map(lambda x: x["date_added"], manifest_resource.get("objects", [])))
    if len(times) > 0:
        headers["X-TAXII-Date-Added-First"] = times[0]
        headers["X-TAXII-Date-Added-Last"] = times[-1]

    return headers


def parse_request_parameters(filter_args):
    """Generates a dict with params received from client"""
    session_args = {}
    for key, value in filter_args.items():
        if key != "limit" and key != "next":
            session_args[key] = set(value.replace(" ", "").split(","))
    return session_args


def find_att(obj):
    """
    Used for finding the version attribute of an ambiguous object. Manifests
    use the "version" field, but objects will use "modified", or if that's not
    available, the "created" field.

    Args:
        obj (dict): manifest or stix object

    Returns:
        string value of the field from the object to use for versioning

    """
    if "version" in obj:
        return string_to_datetime(obj["version"])
    elif "modified" in obj:
        return string_to_datetime(obj["modified"])
    elif "created" in obj:
        return string_to_datetime(obj["created"])
    else:
        return string_to_datetime(obj["_date_added"])


def find_version_attribute(obj):
    """Depending on the object, modified, created or _date_added is used to store the
    object version"""
    if "modified" in obj:
        return "modified"
    elif "created" in obj:
        return "created"
    elif "_date_added" in obj:
        return "_date_added"


class SessionChecker(object):
    """Calls a target method every X seconds to perform a task."""

    def __init__(self, interval, target_function):
        self.interval = interval
        self.target_function = target_function
        self.thread = threading.Timer(interval=self.interval, function=self.handle_function)
        self.thread.daemon = True

    def handle_function(self):
        self.target_function()
        self.thread = threading.Timer(interval=self.interval, function=self.handle_function)
        self.thread.daemon = True
        self.thread.start()

    def start(self):
        self.thread.start()



import re

from flask import request

from ..exceptions import ProcessingError

MEDIA_TYPE_TAXII_ANY = "application/taxii+json"
MEDIA_TYPE_TAXII_V21 = "{media};version=2.1".format(media=MEDIA_TYPE_TAXII_ANY)


def validate_version_parameter_in_accept_header():
    """All endpoints need to check the Accept Header for the correct Media Type"""
    accept_header = request.headers.get("accept", "").replace(" ", "").split(",")
    found = False

    for item in accept_header:
        result = re.match(r"^application/taxii\+json(;version=(\d\.\d))?$", item)
        if result:
            if len(result.groups()) >= 1:
                version_str = result.group(2)
                if version_str != "2.1":  # The server only supports 2.1
                    raise ProcessingError("The server does not support version {}".format(version_str), 406)
            found = True
            break

    if found is False:
        raise ProcessingError("Media type in the Accept header is invalid or not found", 406)



from flask import Blueprint, Response, current_app, json

from . import MEDIA_TYPE_TAXII_V21, validate_version_parameter_in_accept_header
from .. import auth
from .discovery import api_root_exists
from .objects import collection_exists

collections_bp = Blueprint("collections", __name__)


@collections_bp.route("/<string:api_root>/collections/", methods=["GET"])
@auth.login_required
def get_collections(api_root):
    """
    Defines TAXII API - Collections:
    Get Collection section (5.1) `here <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107533>`__

    Args:
        api_root (str): the base URL of the API Root

    Returns:
        collections: A Collections Resource upon successful requests. Additional information
        `here <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107534>`__.

    """
    # TODO: Check if user has access to the each collection's metadata - unrelated to can_read, can_write attributes

    validate_version_parameter_in_accept_header()
    api_root_exists(api_root)
    collections = current_app.medallion_backend.get_collections(api_root)
    return Response(
        response=json.dumps(collections),
        status=200,
        mimetype=MEDIA_TYPE_TAXII_V21,
    )


@collections_bp.route("/<string:api_root>/collections/<string:collection_id>/", methods=["GET"])
@auth.login_required
def get_collection(api_root, collection_id):
    """
    Defines TAXII API - Collections:
    Get Collection section (5.2) `here <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107535>`__

    Args:
        api_root (str): the base URL of the API Root
        collection_id (str): the `identifier` of the Collection being requested

    Returns:
        collection: A Collection Resource upon successful requests. Additional information
        `here <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107536>`__.

    """
    # TODO: Check if user has access to the collection's metadata - unrelated to can_read, can_write attributes

    validate_version_parameter_in_accept_header()
    api_root_exists(api_root)
    collection_exists(api_root, collection_id)
    collection = current_app.medallion_backend.get_collection(api_root, collection_id)

    return Response(
        response=json.dumps(collection),
        status=200,
        mimetype=MEDIA_TYPE_TAXII_V21,
    )



import importlib
import logging

from flask import Flask, Response, current_app, json
from flask_httpauth import HTTPBasicAuth

from .exceptions import BackendError, ProcessingError
from .version import __version__  # noqa
from .views import MEDIA_TYPE_TAXII_V21

# Console Handler for medallion messages
ch = logging.StreamHandler()
ch.setFormatter(logging.Formatter("[%(name)s] [%(levelname)-8s] [%(asctime)s] %(message)s"))

# Module-level logger
log = logging.getLogger(__name__)
log.addHandler(ch)

application_instance = Flask(__name__)
auth = HTTPBasicAuth()


def load_app(config_file):
    with open(config_file, "r") as f:
        configuration = json.load(f)

    set_config(application_instance, "users", configuration)
    set_config(application_instance, "taxii", configuration)
    set_config(application_instance, "backend", configuration)
    register_blueprints(application_instance)

    return application_instance


def set_config(flask_application_instance, prop_name, config):
    with flask_application_instance.app_context():
        log.debug("Registering medallion {} configuration into {}".format(prop_name, current_app))
        if prop_name == "taxii":
            try:
                flask_application_instance.taxii_config = config[prop_name]
            except KeyError:
                flask_application_instance.taxii_config = {'max_page_size': 100}
        elif prop_name == "users":
            try:
                flask_application_instance.users_backend = config[prop_name]
            except KeyError:
                log.warning("You did not give user information in your config.")
                log.warning("We are giving you the default user information of:")
                log.warning("User = user")
                log.warning("Pass = pass")
                flask_application_instance.users_backend = {"user": "pass"}
        elif prop_name == "backend":
            try:
                flask_application_instance.medallion_backend = connect_to_backend(config[prop_name])
            except KeyError:
                log.warning("You did not give backend information in your config.")
                log.warning("We are giving medallion the default settings,")
                log.warning("which includes a data file of 'default_data.json'.")
                log.warning("Please ensure this file is in your CWD.")
                back = {'module': 'medallion.backends.memory_backend', 'module_class': 'MemoryBackend', 'filename': None}
                flask_application_instance.medallion_backend = connect_to_backend(back)


def connect_to_backend(config_info):
    log.debug("Initializing backend configuration using: {}".format(config_info))

    if "module" not in config_info:
        raise ValueError("No module parameter provided for the TAXII server.")
    if "module_class" not in config_info:
        raise ValueError("No module_class parameter provided for the TAXII server.")

    try:
        module = importlib.import_module(config_info["module"])
        module_class = getattr(module, config_info["module_class"])
        log.debug("Instantiating medallion backend with {}".format(module_class))
        return module_class(**config_info)
    except Exception as e:
        log.error("Unknown backend for TAXII server. {} ".format(str(e)))
        raise


def register_blueprints(flask_application_instance):
    from medallion.views import collections
    from medallion.views import discovery
    from medallion.views import manifest
    from medallion.views import objects

    with flask_application_instance.app_context():
        log.debug("Registering medallion blueprints into {}".format(current_app))
        current_app.register_blueprint(collections.collections_bp)
        current_app.register_blueprint(discovery.discovery_bp)
        current_app.register_blueprint(manifest.manifest_bp)
        current_app.register_blueprint(objects.objects_bp)


@auth.get_password
def get_pwd(username):
    if username in current_app.users_backend:
        return current_app.users_backend.get(username)
    return None


@application_instance.errorhandler(500)
def handle_error(error):
    e = {
        "title": "InternalError",
        "http_status": "500",
        "description": str(error.args[0]),
    }
    return Response(
        response=json.dumps(e),
        status=500,
        mimetype=MEDIA_TYPE_TAXII_V21,
    )


@application_instance.errorhandler(ProcessingError)
def handle_processing_error(error):
    e = {
        "title": str(error.__class__.__name__),
        "http_status": str(error.status),
        "description": str(error),
    }
    return Response(
        response=json.dumps(e),
        status=error.status,
        headers=getattr(error, "headers", None),
        mimetype=MEDIA_TYPE_TAXII_V21,
    )


@application_instance.errorhandler(BackendError)
def handle_backend_error(error):
    e = {
        "title": str(error.__class__.__name__),
        "http_status": str(error.status),
        "description": str(error),
    }
    return Response(
        response=json.dumps(e),
        status=error.status,
        mimetype=MEDIA_TYPE_TAXII_V21,
    )



import logging
import re

from flask import Blueprint, Response, current_app, json, request

from . import MEDIA_TYPE_TAXII_V21, validate_version_parameter_in_accept_header
from .. import auth
from ..common import get_timestamp
from ..exceptions import ProcessingError
from .discovery import api_root_exists

objects_bp = Blueprint("objects", __name__)

# Module-level logger
log = logging.getLogger(__name__)


def permission_to_read(api_root, collection_id):
    collection_info = current_app.medallion_backend.get_collection(api_root, collection_id)
    if collection_info["can_read"] is False:
        raise ProcessingError("Forbidden to read collection '{}'".format(collection_id), 403)


def permission_to_write(api_root, collection_id):
    collection_info = current_app.medallion_backend.get_collection(api_root, collection_id)
    if collection_info["can_write"] is False:
        raise ProcessingError("Forbidden to write collection '{}'".format(collection_id), 403)


def permission_to_read_and_write(api_root, collection_id):
    collection_info = current_app.medallion_backend.get_collection(api_root, collection_id)
    if collection_info["can_read"] is False and collection_info["can_write"] is False:
        raise ProcessingError("Collection '{}' not found".format(collection_id), 404)
    if collection_info["can_write"] is False:
        raise ProcessingError("Forbidden to write collection '{}'".format(collection_id), 403)
    if collection_info["can_read"] is False:
        raise ProcessingError("Forbidden to read collection '{}'".format(collection_id), 403)


def collection_exists(api_root, collection_id):
    if not current_app.medallion_backend.get_collection(api_root, collection_id):
        raise ProcessingError("Collection '{}' not found".format(collection_id), 404)


def validate_size_in_request_body(api_root):
    api_root = current_app.medallion_backend.get_api_root_information(api_root)
    max_length = api_root["max_content_length"]
    try:
        content_length = int(request.headers.get("content_length", ""))
    except ValueError:
        raise ProcessingError("The server did not understand the request or headers", 400)
    if content_length > max_length or content_length <= 0:
        raise ProcessingError("Content-Length header not valid or exceeds maximum!", 413)


def validate_version_parameter_in_content_type_header():
    content_type = request.headers.get("content_type", "").replace(" ", "").split(",")
    found = False

    for item in content_type:
        result = re.match(r"^application/taxii\+json(;version=(\d\.\d))?$", item)
        if result:
            if len(result.groups()) >= 2:
                version_str = result.group(2)
                if version_str != "2.1":  # The server only supports 2.1
                    raise ProcessingError("The server does not support version {}".format(version_str), 415)
            found = True
            break

    if found is False:
        raise ProcessingError("Media type in the Content-Type header is invalid or not found", 415)


def validate_limit_parameter():
    max_page = current_app.taxii_config["max_page_size"]
    limit = request.args.get("limit", max_page)
    try:
        limit = int(limit)
    except ValueError:
        raise ProcessingError("The server did not understand the request or filter parameters", 400)
    if limit <= 0:
        raise ProcessingError("The limit parameter cannot be negative or zero", 400)
    if limit > max_page:
        limit = max_page
    return limit


@objects_bp.route("/<string:api_root>/collections/<string:collection_id>/objects/", methods=["GET", "POST"])
@auth.login_required
def get_or_add_objects(api_root, collection_id):
    """
    Defines TAXII API - Collections:
        Get Objects section (`5.4 <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107539>`__)
        and Add Objects section (`5.5 <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107540>`__)

    Args:
        api_root (str): the base URL of the API Root
        collection_id (str): the `identifier` of the Collection being requested

    Returns:
        resource:
            GET -> An Envelope Resource upon successful requests.
            POST -> An Status Resource upon successful requests.

    """
    # TODO: Check if user has access to read or write objects in collection - right now just check for permissions on the collection.
    request_time = get_timestamp()  # Can't I get this from the request itself?
    validate_version_parameter_in_accept_header()
    api_root_exists(api_root)
    collection_exists(api_root, collection_id)

    if request.method == "GET":
        permission_to_read(api_root, collection_id)
        limit = validate_limit_parameter()
        objects, headers = current_app.medallion_backend.get_objects(
            api_root, collection_id, request.args.to_dict(), ("id", "type", "version", "spec_version"), limit
        )

        return Response(
            response=json.dumps(objects),
            status=200,
            headers=headers,
            mimetype=MEDIA_TYPE_TAXII_V21,
        )

    elif request.method == "POST":
        validate_version_parameter_in_content_type_header()
        permission_to_write(api_root, collection_id)
        validate_size_in_request_body(api_root)
        status = current_app.medallion_backend.add_objects(
            api_root, collection_id, request.get_json(force=True), request_time
        )
        return Response(
            response=json.dumps(status),
            status=202,
            mimetype=MEDIA_TYPE_TAXII_V21,
        )


@objects_bp.route("/<string:api_root>/collections/<string:collection_id>/objects/<string:object_id>/", methods=["GET", "DELETE"])
@auth.login_required
def get_or_delete_object(api_root, collection_id, object_id):
    """
    Defines TAXII API - Collections:
        Get Object section (`5.6 <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107541>`__)
        and Delete Object section (`5.7 <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107542>`__)

    Args:
        api_root (str): the base URL of the API Root
        collection_id (str): the `identifier` of the Collection being requested
        object_id (str): the `identifier` of the object being requested

    Returns:
        resource:
            GET -> An Envelope Resource upon successful requests.
            DELETE -> Upon successful request nothing (status code 200).

    """
    # TODO: Check if user has access to read or write objects in collection - right now just check for permissions on the collection.
    validate_version_parameter_in_accept_header()
    api_root_exists(api_root)
    collection_exists(api_root, collection_id)

    if request.method == "GET":
        permission_to_read(api_root, collection_id)
        limit = validate_limit_parameter()
        objects, headers = current_app.medallion_backend.get_object(
            api_root, collection_id, object_id, request.args.to_dict(), ("version", "spec_version"), limit
        )
        if objects or request.args:
            return Response(
                response=json.dumps(objects),
                status=200,
                headers=headers,
                mimetype=MEDIA_TYPE_TAXII_V21,
            )
        raise ProcessingError("Object '{}' not found".format(object_id), 404)
    elif request.method == "DELETE":
        permission_to_read_and_write(api_root, collection_id)
        current_app.medallion_backend.delete_object(
            api_root, collection_id, object_id, request.args.to_dict(), ("version", "spec_version"),
        )
        return Response(
            status=200,
            mimetype=MEDIA_TYPE_TAXII_V21,
        )


@objects_bp.route("/<string:api_root>/collections/<string:collection_id>/objects/<string:object_id>/versions/", methods=["GET"])
@auth.login_required
def get_object_versions(api_root, collection_id, object_id):
    """
    Defines TAXII API - Collections: Get Object Versions section
    `(5.8) <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107543>`__.

    Args:
        api_root (str): the base URL of the API Root
        collection_id (str): the `identifier` of the Collection being requested
        object_id (str): the `identifier` of the object being requested

    Returns:
        versions: A Versions Resource upon successful requests. Additional information
            `here <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107544>`__.

    """
    # TODO: Check if user has access to read objects in collection - right now just check for permissions on the collection.

    validate_version_parameter_in_accept_header()
    api_root_exists(api_root)
    collection_exists(api_root, collection_id)
    permission_to_read(api_root, collection_id)

    limit = validate_limit_parameter()
    versions, headers = current_app.medallion_backend.get_object_versions(
        api_root, collection_id, object_id, request.args.to_dict(), ("spec_version",), limit
    )
    return Response(
        response=json.dumps(versions),
        status=200,
        headers=headers,
        mimetype=MEDIA_TYPE_TAXII_V21,
    )



#!/usr/bin/python
# -*- coding: utf-8 -*-

import argparse
import codecs
import copy
import json
import multiprocessing.pool
import os
import shutil
import textwrap
import zipfile

import six

import requests
from taxii2client import v20, v21

__version__ = "1.0"
__author__ = "emmanvg"


def download_file(item):
    """Download file based on the given url, save to disk with the provided filename."""
    filename, url = item
    if os.path.isfile(filename) is False:
        print("Downloading {}...".format(url))
        try:
            response = requests.get(url, stream=True)
            if response.status_code == 200:
                with open(filename, mode="wb") as outfile:
                    for chunk in response:
                        outfile.write(chunk)
                print("Finished saving file '{}'".format(filename))
            else:
                response.raise_for_status()
        except requests.exceptions.SSLError as e:
            print("Please check if you are behind a proxy, request was unsuccessful...")
            print(e)
    else:
        print("File {} is already present...".format(filename))
    return filename


def process_json_object(collection, filename):
    """Open the file and send it to the server. Save the status resource returned by the server to disk."""
    with codecs.open(filename, mode="r", encoding="utf8") as infile:
        status = collection.add_objects(json.load(infile), timeout=0)
        status = status._raw

    status_filename = status["id"] + ".json"
    with codecs.open(status_filename, mode="w", encoding="utf8") as out_status:
        output = json.dumps(status, ensure_ascii=False, indent=4, separators=(',', ': '))
        out_status.write(six.text_type(output))
    print("Wrote status response in '{}' for {}".format(status_filename, filename))
    return status_filename


def post_to_taxii_server(item):
    """Method called by each item in the web_locations list in preparation to be added to server."""
    filename, url, collection, delete_download, delete_status = item
    print("Adding objects from '{}'...".format(filename))

    if zipfile.is_zipfile(filename):
        z = zipfile.ZipFile(filename)
        status_filenames = []
        for zip_member in z.namelist():
            if zip_member.endswith(".json"):
                location = z.extract(member=zip_member)
                status = process_json_object(collection, location)
                status_filenames.append(status)
        if delete_status:
            for s in status_filenames:
                remove_file(s)
        if delete_download:
            for item in sorted(z.NameToInfo.keys(), reverse=True):
                remove_file(item)
        z.close()
        if delete_download:
            remove_file(filename)

    elif filename.endswith(".json"):
        status = process_json_object(collection, filename)
        if delete_status:
            remove_file(status)
        if delete_download:
            remove_file(filename)


def print_json_message(message):
    """Used in the interactive session to print messages received from TAXII Server."""
    print("{:*^50}".format("RESPONSE BEGIN"))
    print(json.dumps(message, indent=4))
    print("{:*^50}".format("RESPONSE END"))


def get_collection_for_post(url, username, password, version):
    """Interactive session to select an available collection to insert the downloaded
    data to."""
    if version == "2.1":
        server = v21.Server(url + "taxii2/", user=username, password=password)
    else:
        server = v20.Server(url + "taxii/", user=username, password=password)
    print_json_message(server._raw)

    count_apiroot = len(server.api_roots)
    if count_apiroot > 0:
        select_apiroot = input("Select an ApiRoot (1 - {}): ".format(count_apiroot))
        api_root = server.api_roots[int(select_apiroot) - 1]
    else:
        raise RuntimeError("No ApiRoots available...")

    count_collections = len(api_root.collections)
    if count_collections > 0:
        print_json_message({"collections": [x._raw for x in api_root.collections]})
        select_collection = input("Select Collection to POST (1 - {}): ".format(count_collections))
        collection = api_root.collections[int(select_collection) - 1]
    else:
        raise RuntimeError("No collections available...")

    return collection


def remove_file(path):
    """Check if the file exists, then remove it. Works only on the current directory where"""
    full_path = os.path.join(os.getcwd(), path)
    if os.path.isfile(full_path):
        os.remove(full_path)
        print("Deleted file {}...".format(full_path))
    elif os.path.isdir(full_path):
        shutil.rmtree(full_path, ignore_errors=True)
        print("Deleted directory {}...".format(full_path))


def _add_certs():
    """Helper method to add certs into certifi package. This method is not called by the script"""
    import certifi
    certifi_location = certifi.where()
    my_certs = "example.crt"  # update this filename as required

    with open(my_certs, "rb") as infile:
        customca = infile.read()

    # append the new certs to the
    with open(certifi_location, "ab") as outfile:
        outfile.write(customca)


class NewlinesHelpFormatter(argparse.RawDescriptionHelpFormatter):
    """Custom help formatter to insert newlines between argument help texts.
    """
    def _split_lines(self, text, width):
        text = self._whitespace_matcher.sub(" ", text).strip()
        txt = textwrap.wrap(text, width)
        txt[-1] += "\n"
        return txt


def _get_argparser():
    """Create and return an ``ArgumentParser`` instance for this application."""
    desc = "TAXII POST-ing script for bulk additions to a specific collection v{}".format(__version__)
    parser = argparse.ArgumentParser(
        description=desc,
        formatter_class=NewlinesHelpFormatter,
    )

    parser.add_argument(
        "-l",
        "--host",
        default="http://localhost:5000/",
        action="store",
        type=str,
        help="<url to TAXII Server or specific collection>"
    )

    parser.add_argument(
        "-u",
        "--user",
        default="admin",
        action="store",
        type=str,
        help="<username for basic auth>"
    )

    parser.add_argument(
        "-p",
        "--password",
        default="Password0",
        action="store",
        type=str,
        help="<password for basic auth>"
    )

    parser.add_argument(
        "--request-version",
        default="2.1",
        action="store",
        type=str,
        help="<client version to perform the request>"
    )

    parser.add_argument(
        "-w",
        "--web-locations",
        nargs="+",
        type=str,
        action="append",
        help="<web locations to download from> example: -w filename.json https://example.com/file.json -w filename2.json https://example2.com/file.json"
    )

    parser.add_argument(
        "--clean-downloads",
        default=False,
        action="store_true",
        help="<when flag is set, removes downloaded files>"
    )

    parser.add_argument(
        "--clean-statuses",
        default=False,
        action="store_true",
        help="<when flag is set, removes the status json files received from server>"
    )

    return parser


def main():
    parser = _get_argparser()
    tool_args = parser.parse_args()

    try:
        # v21.Connection()
        if tool_args.request_version == "2.1":
            selected_collection = v21.Collection(tool_args.host, user=tool_args.user, password=tool_args.password)
        else:
            selected_collection = v20.Collection(tool_args.host, user=tool_args.user, password=tool_args.password)
        selected_collection.refresh()
    except requests.HTTPError:
        selected_collection = get_collection_for_post(tool_args.host, tool_args.user, tool_args.password, tool_args.request_version)

    # these urls and filenames are used for downloading and populating the medallion backend
    web_locations = [
        ["enterprise-attack.json", "https://github.com/mitre/cti/raw/master/enterprise-attack/enterprise-attack.json"],
        ["mobile-attack.json", "https://github.com/mitre/cti/raw/master/mobile-attack/mobile-attack.json"],
        ["pre-attack.json", "https://github.com/mitre/cti/raw/master/pre-attack/pre-attack.json"],
        ["stix-capec.json", "https://github.com/mitre/cti/raw/master/capec/stix-capec.json"],
        ["cti-stix-elevator.zip", "https://github.com/oasis-open/cti-stix-elevator/archive/master.zip"],
    ]

    with multiprocessing.pool.ThreadPool() as threads:
        threads.imap_unordered(download_file, web_locations)
        threads.close()
        threads.join()

    for loc in web_locations:
        loc.append(copy.deepcopy(selected_collection))
        loc.append(tool_args.clean_downloads)
        loc.append(tool_args.clean_statuses)

    with multiprocessing.pool.ThreadPool() as threads:
        threads.imap_unordered(post_to_taxii_server, web_locations)
        threads.close()
        threads.join()


if __name__ == "__main__":
    main()



import calendar
import datetime

import pytz

from medallion.common import datetime_to_float, string_to_datetime

if __name__ == '__main__':
    timestamp_str = "2019-11-24T23:29:07.434659Z"

    epoch_naive = datetime.datetime(1970, 1, 1, 0, 0, 0, 0)
    epoch_utc = epoch_naive.replace(tzinfo=pytz.UTC)

    dt_naive = string_to_datetime(timestamp_str)
    dt_utc = dt_naive.replace(tzinfo=pytz.UTC)

    diff_naive = (dt_naive - epoch_naive).total_seconds()
    diff_utc = (dt_utc - epoch_utc).total_seconds()
    diff_medallion = datetime_to_float(dt_naive)
    diff_calendar = calendar.timegm(dt_naive.utctimetuple())

    print("Naive epoch difference:\t", diff_naive)
    print("UTC epoch difference:\t", diff_utc)
    print("Medallion difference:\t", diff_medallion)
    print("Via calendar.timegm():\t", diff_calendar)




class Backend(object):

    def server_discovery(self):
        """
        Fill:
            Returns the discovery information (api_roots, etc) for this server

        Args:
            -none-

        Returns:
            discovery information

        """
        raise NotImplementedError()

    def get_collections(self, api_root):
        """
        Fill:
            Implement the get_collections TAXII endpoint by obtaining the collection metadata
            for this api_root

        Args:
            api_root (str): the name of the api_root.

        Returns:
            tuple containing the total count, and metadata for all collections at this api root

        """
        raise NotImplementedError()

    def get_collection(self, api_root, collection_id):
        """
        Fill:
            Implement the get_collection TAXII service by obtaining the collection metadata
            for collection

        Args:
            api_root (str): the name of the api_root.
            collection_id (str): the id of the collection

        Returns:
            collection metadata

        """
        raise NotImplementedError()

    def get_object_manifest(self, api_root, collection_id, filter_args, allowed_filters, limit):
        """
        Fill:
            Implement the get_object_manifest TAXII endpoint by obtaining the metadata
            for the selected objects

        Args:
            api_root (str): the name of the api_root.
            collection_id (str): the id of the collection
            filter_args (werkzeug.datastructures.ImmutableMultiDict): query string from URL
                containing filter args
            allowed_filters (tuple): STIX properties which are allowed in the filter for this endpoint
            limit (int): Used for pagination requests. limits objects to the amount specified

        Returns:
            tuple containing the total count of matching objects, and a collection of metadata for the objects

        """
        raise NotImplementedError()

    def get_api_root_information(self, api_root):
        """
        Fill:
            Implement the get_api_root_information TAXII endpoint by obtaining api_root metadata

        Args:
            api_root (str): the name of the api_root.

        Returns:
            metadata for the api root

        """
        raise NotImplementedError()

    def get_status(self, api_root, status_id):
        """
        Fill:
            Implement the get_status TAXII endpoint by obtaining the status of an add_objects request

        Args:
            api_root (str): the name of the api_root.
            status_id (str): the id of the add_objects request

        Returns:
            status of the request (including):
                how many objects were successful saved
                how many objects failed to be saved
                how many objects are pending

        """
        raise NotImplementedError()

    def get_objects(self, api_root, collection_id, filter_args, allowed_filters, limit):
        """
        Fill:
            Implement the get_objects TAXII endpoint by obtaining the data from a collection

        Args:
            api_root (str): the name of the api_root.
            collection_id (str): the id of the collection
            filter_args (werkzeug.datastructures.ImmutableMultiDict): query string from URL
                containing filter args
            allowed_filters (tuple): STIX properties which are allowed in the filter for this endpoint
            limit (int): Used for pagination requests. limits objects to the amount specified

        Returns:
            tuple containing the total count of matching objects, and a collection of objects containing
            the data from the collection that satisfies the filter

        """
        raise NotImplementedError()

    def add_objects(self, api_root, collection_id, objs, request_time):
        """
        Fill:
            Implement the add_objects TAXII endpoint by save data into a collection

        Args:
            api_root (str): the name of the api_root.
            collection_id (str): the id of the collection
            objs (dict): bundle containing objects to insert into the collection
            request_time (str): a formatted timestamp string with the time of the request

        Returns:
            status of the request (including):
                how many objects were successful saved
                how many objects failed to be saved
                how many objects are pending

        METADATA ABOUT EACH SUCCESSFUL OBJECT SAVED MUST BE AVAILABLE VIA THE get_object_manifest API CALL
        THIS CAN BE IMPLEMENTED AS A SEPARATE STORE, OTHERWISE IT NEEDS TO BE GENERATABLE DYNAMICALLY

        """
        raise NotImplementedError()

    def get_object(self, api_root, collection_id, object_id, filter_args, allowed_filters, limit):
        """
        Fill:
            Implement the get_object TAXII endpoint by obtaining the data from a collection related
            to the object_id

        Args:
            api_root (str): the name of the api_root.
            collection_id (str): the id of the collection
            object_id (str): the id of the requested object
            filter_args (werkzeug.datastructures.ImmutableMultiDict): query string from URL
                containing filter args
            allowed_filters (tuple): STIX properties which are allowed in the filter for this endpoint
            limit (int): Used for pagination requests. limits objects to the amount specified

        Returns:
            data from the collection that satisfies the filter

        """
        raise NotImplementedError()

    def delete_object(self, api_root, collection_id, object_id, filter_args, allowed_filters):
        """
        Fill:
            Implement the delete_object TAXII endpoint by obtaining the metadata for a selected
            object

        Args:
            api_root (str): the name of the api_root.
            collection_id (str): the id of the collection
            object_id (str): the id of the requested object
            filter_args (werkzeug.datastructures.ImmutableMultiDict): query string from URL
                containing filter args
            allowed_filters (tuple): STIX properties which are allowed in the filter for this endpoint

        Returns:
            Nothing.

        """
        raise NotImplementedError()

    def get_object_versions(self, api_root, collection_id, object_id, filter_args, allowed_filters, limit):
        """
        Fill:
            Implement the get_object_versions TAXII endpoint by obtaining the metadata for a selected
            object

        Args:
            api_root (str): the name of the api_root.
            collection_id (str): the id of the collection
            object_id (str): the id of the requested object
            filter_args (werkzeug.datastructures.ImmutableMultiDict): query string from URL
                containing filter args
            allowed_filters (tuple): STIX properties which are allowed in the filter for this endpoint
            limit (int): Used for pagination requests. limits objects to the amount specified

        Returns:
            data from the collection that satisfies the filter

        """
        raise NotImplementedError()



import bisect
import copy
import operator

from ..common import find_att, string_to_datetime


def check_for_dupes(final_match, final_track, res):
    for obj in res:
        found = 0
        pos = bisect.bisect_left(final_track, obj["id"])
        if not final_match or pos > len(final_track) - 1 or final_track[pos] != obj["id"]:
            final_track.insert(pos, obj["id"])
            final_match.insert(pos, obj)
        else:
            obj_time = find_att(obj)
            while pos != len(final_track) and obj["id"] == final_track[pos]:
                if find_att(final_match[pos]) == obj_time:
                    found = 1
                    break
                else:
                    pos = pos + 1
            if found == 1:
                continue
            else:
                final_track.insert(pos, obj["id"])
                final_match.insert(pos, obj)


def check_version(data, relate):
    id_track = []
    res = []
    for obj in data:
        pos = bisect.bisect_left(id_track, obj["id"])
        if not res or pos >= len(id_track) or id_track[pos] != obj["id"]:
            id_track.insert(pos, obj["id"])
            res.insert(pos, obj)
        else:
            if relate(find_att(obj), find_att(res[pos])):
                res[pos] = obj
    return res


class BasicFilter(object):

    def __init__(self, filter_args):
        self.filter_args = filter_args

    def sort_and_paginate(self, data, limit, manifest):
        temp = None
        next_save = {}
        headers = {}
        new = []
        if len(data) == 0:
            return new, next_save, headers
        if manifest:
            manifest.sort(key=lambda x: x['date_added'])
            for man in manifest:
                man_time = find_att(man)
                for check in data:
                    check_time = find_att(check)
                    if check['id'] == man['id'] and check_time == man_time:
                        if len(headers) == 0:
                            headers["X-TAXII-Date-Added-First"] = man["date_added"]
                        new.append(check)
                        temp = man
                        if len(new) == limit:
                            headers["X-TAXII-Date-Added-Last"] = man["date_added"]
                        break
            if limit and limit < len(data):
                next_save = new[limit:]
                new = new[:limit]
            else:
                headers["X-TAXII-Date-Added-Last"] = temp["date_added"]
        else:
            data.sort(key=lambda x: x['date_added'])
            if limit and limit < len(data):
                next_save = data[limit:]
                data = data[:limit]
            headers["X-TAXII-Date-Added-First"] = data[0]["date_added"]
            headers["X-TAXII-Date-Added-Last"] = data[-1]["date_added"]
            new = data
        return new, next_save, headers

    @staticmethod
    def filter_by_id(data, id_):
        id_ = id_.split(",")

        match_objects = []

        for obj in data:
            if "id" in obj and any(s == obj["id"] for s in id_):
                match_objects.append(obj)

        return match_objects

    @staticmethod
    def filter_by_added_after(data, manifest_info, added_after_date):
        added_after_timestamp = string_to_datetime(added_after_date)
        new_results = []
        # for manifest objects and versions
        if manifest_info is None:
            for obj in data:
                if string_to_datetime(obj["date_added"]) > added_after_timestamp:
                    new_results.append(obj)
        # for other objects with manifests
        else:
            for obj in data:
                obj_time = find_att(obj)
                for item in manifest_info:
                    item_time = find_att(item)
                    if item["id"] == obj["id"] and item_time == obj_time and string_to_datetime(item["date_added"]) > added_after_timestamp:
                        new_results.append(obj)
                        break
        return new_results

    @staticmethod
    def filter_by_version(data, version):
        # final_match is a sorted list of objects
        final_match = []
        # final_track is a sorted list of id's
        final_track = []

        # return most recent object versions unless otherwise specified
        if version is None:
            version = "last"
        version_indicators = version.split(",")

        if "all" in version_indicators:
            # if "all" is in the list, just return everything
            return data

        actual_dates = [string_to_datetime(x) for x in version_indicators if x != "first" and x != "last"]
        # if a specific version is given, filter for objects with that value
        if actual_dates:
            id_track = []
            res = []
            for obj in data:
                obj_time = find_att(obj)
                if obj_time in actual_dates:
                    pos = bisect.bisect_left(id_track, obj["id"])
                    id_track.insert(pos, obj["id"])
                    res.insert(pos, obj)
            final_match = res
            final_track = id_track

        if "first" in version_indicators:
            res = check_version(data, operator.lt)
            check_for_dupes(final_match, final_track, res)

        if "last" in version_indicators:
            res = check_version(data, operator.gt)
            check_for_dupes(final_match, final_track, res)

        return final_match

    @staticmethod
    def filter_by_type(data, type_):
        type_ = type_.split(",")
        match_objects = []

        for obj in data:
            if "type" in obj and any(s == obj["type"] for s in type_):
                match_objects.append(obj)
            elif "id" in obj and any(s == obj["id"].split("--")[0] for s in type_):
                match_objects.append(obj)

        return match_objects

    @staticmethod
    def filter_by_spec_version(data, spec_):
        spec_ = spec_.split(",")

        match_objects = []

        for obj in data:
            if "spec_version" in obj and any(s == obj["spec_version"] for s in spec_):
                match_objects.append(obj)
            elif "media_type" in obj and any(s == obj["media_type"].split("version=")[1] for s in spec_):
                match_objects.append(obj)

        return match_objects

    def process_filter(self, data, allowed, manifest_info, limit):
        filtered_by_type = []
        filtered_by_id = []
        filtered_by_spec_version = []
        filtered_by_added_after = []
        filtered_by_version = []
        final_match = []
        save_next = []
        headers = {}

        # match for type and id filters first
        match_type = self.filter_args.get("match[type]")
        if match_type and "type" in allowed:
            filtered_by_type = self.filter_by_type(data, match_type)
        else:
            filtered_by_type = copy.deepcopy(data)

        match_id = self.filter_args.get("match[id]")
        if match_id and "id" in allowed:
            filtered_by_id = self.filter_by_id(filtered_by_type, match_id)
        else:
            filtered_by_id = filtered_by_type

        # match for spec_version
        match_spec_version = self.filter_args.get("match[spec_version]")
        if match_spec_version and "spec_version" in allowed:
            filtered_by_spec_version = self.filter_by_spec_version(filtered_by_id, match_spec_version)
        else:
            filtered_by_spec_version = filtered_by_id

        # match for added_after
        added_after_date = self.filter_args.get("added_after")
        if added_after_date:
            filtered_by_added_after = self.filter_by_added_after(filtered_by_spec_version, manifest_info, added_after_date)
        else:
            filtered_by_added_after = filtered_by_spec_version

        # match for version, and get rid of duplicates as appropriate
        if "version" in allowed:
            match_version = self.filter_args.get("match[version]")
            filtered_by_version = self.filter_by_version(filtered_by_added_after, match_version)
        else:
            filtered_by_version = filtered_by_added_after

        # sort objects by date_added of manifest and paginate as necessary
        final_match, save_next, headers = self.sort_and_paginate(filtered_by_version, limit, manifest_info)

        return final_match, save_next, headers



#!/usr/bin/env python
from codecs import open
import os.path

from setuptools import find_packages, setup

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
VERSION_FILE = os.path.join(BASE_DIR, "medallion", "version.py")


def get_version():
    with open(VERSION_FILE, encoding="utf-8") as f:
        for line in f.readlines():
            if line.startswith("__version__"):
                version = line.split()[-1].strip("\"")
                return version
        raise AttributeError("Package does not have a __version__")


def get_long_description():
    with open("README.rst", encoding="utf-8") as f:
        return f.read()


setup(
    name="medallion",
    version=get_version(),
    description="A TAXII 2.1 Server implementing required endpoints",
    long_description=get_long_description(),
    long_description_content_type="text/x-rst",
    url="https://oasis-open.github.io/cti-documentation/",
    author="OASIS Cyber Threat Intelligence Technical Committee",
    author_email="cti-users@lists.oasis-open.org",
    maintainer="Emmanuelle Vargas-Gonzalez",
    maintainer_email="emmanuelle@mitre.org",
    license="BSD",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Security",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="taxii taxii2 server json cti cyber threat intelligence",
    packages=find_packages(exclude=["*.test", "*.test.data"]),
    install_requires=[
        "flask>=0.12.1",
        "Flask-HTTPAuth",
        "pytz",
        "six",
    ],
    entry_points={
        "console_scripts": [
            "medallion = medallion.scripts.run:main",
        ],
    },
    extras_require={
        "test": [
            "coverage",
            "pytest",
            "pytest-cov",
            "tox",
        ],
        "docs": [
            "sphinx",
            "sphinx-prompt",
        ],
        "mongo": [
            "pymongo",
        ],
    },
    project_urls={
        'Documentation': 'https://medallion.readthedocs.io/',
        'Source Code': 'https://github.com/oasis-open/cti-taxii-server/',
        'Bug Tracker': 'https://github.com/oasis-open/cti-taxii-server/issues/',
    },
)



from flask import Blueprint, Response, current_app, json, request

from . import MEDIA_TYPE_TAXII_V21, validate_version_parameter_in_accept_header
from .. import auth
from .discovery import api_root_exists
from .objects import (
    collection_exists, permission_to_read, validate_limit_parameter
)

manifest_bp = Blueprint("manifest", __name__)


@manifest_bp.route("/<string:api_root>/collections/<string:collection_id>/manifest/", methods=["GET"])
@auth.login_required
def get_object_manifest(api_root, collection_id):
    """
    Defines TAXII API - Collections:
    Get Object Manifests section (5.3) `here <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107537>`__

    Args:
        api_root (str): the base URL of the API Root
        collection_id (str): the `identifier` of the Collection being requested

    Returns:
        manifest: A Manifest Resource upon successful requests. Additional information
        `here <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107538>`__.

    """
    validate_version_parameter_in_accept_header()
    api_root_exists(api_root)
    collection_exists(api_root, collection_id)
    permission_to_read(api_root, collection_id)

    limit = validate_limit_parameter()
    manifests, headers = current_app.medallion_backend.get_object_manifest(
        api_root, collection_id, request.args.to_dict(), ("id", "type", "version", "spec_version"), limit
    )

    return Response(
        response=json.dumps(manifests),
        status=200,
        headers=headers,
        mimetype=MEDIA_TYPE_TAXII_V21,
    )



from flask import Blueprint, Response, current_app, json

from . import MEDIA_TYPE_TAXII_V21, validate_version_parameter_in_accept_header
from .. import auth
from ..exceptions import ProcessingError

discovery_bp = Blueprint("discovery", __name__)


def api_root_exists(api_root):
    result = current_app.medallion_backend.get_api_root_information(api_root)
    if not result:
        raise ProcessingError("API root '{}' information not found".format(api_root), 404)


@discovery_bp.route("/taxii2/", methods=["GET"])
@auth.login_required
def get_server_discovery():
    """
    Defines TAXII API - Server Information:
    Server Discovery section (4.1) `here <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107526>`__

    Returns:
        discovery: A Discovery Resource upon successful requests. Additional information
        `here <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107527>`__.

    """
    # Having access to the discovery method is only related to having
    # credentials on the server. The metadata returned might be different
    # depending upon the credentials.
    validate_version_parameter_in_accept_header()
    server_discovery = current_app.medallion_backend.server_discovery()

    if server_discovery:
        return Response(
            response=json.dumps(server_discovery),
            status=200,
            mimetype=MEDIA_TYPE_TAXII_V21,
        )
    raise ProcessingError("Server discovery information not available", 404)


@discovery_bp.route("/<string:api_root>/", methods=["GET"])
@auth.login_required
def get_api_root_information(api_root):
    """
    Defines TAXII API - Server Information:
    Get API Root Information section (4.2) `here <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107528>`__

    Args:
        api_root (str): the base URL of the API Root

    Returns:
        api-root: An API Root Resource upon successful requests. Additional information
        `here <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107529>`__.

    """
    # TODO: Check if user has access to objects in collection.
    validate_version_parameter_in_accept_header()
    api_root_exists(api_root)
    root_info = current_app.medallion_backend.get_api_root_information(api_root)
    return Response(
        response=json.dumps(root_info),
        status=200,
        mimetype=MEDIA_TYPE_TAXII_V21,
    )


@discovery_bp.route("/<string:api_root>/status/<string:status_id>/", methods=["GET"])
@auth.login_required
def get_status(api_root, status_id):
    """
    Defines TAXII API - Server Information:
    Get Status section (4.3) `here <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107530>`__

    Args:
        api_root (str): the base URL of the API Root
        status_id (str): the `identifier` of the Status message being requested

    Returns:
        status: A Status Resource upon successful requests. Additional information
        `here <https://docs.oasis-open.org/cti/taxii/v2.1/cs01/taxii-v2.1-cs01.html#_Toc31107531>`__.

    """
    # TODO: Check if user has access to the Status resource.
    validate_version_parameter_in_accept_header()
    api_root_exists(api_root)
    status = current_app.medallion_backend.get_status(api_root, status_id)

    if status:
        return Response(
            response=json.dumps(status),
            status=200,
            mimetype=MEDIA_TYPE_TAXII_V21,
        )
    raise ProcessingError("Status '{}' not found".format(status_id), 404)
